/*************************************************************************
 *
 * asset-ident.c
 *
 * Matt Shelton <matt@mattshelton.com>
 *
 * This module contains functions related to identifying an asset.
 *
 * Copyright (C) 2004 Matt Shelton <matt@mattshelton.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: asset-ident.c,v 1.18 2004/09/30 22:25:50 mattshelton Exp $
 *
 **************************************************************************/
#include "asset-ident.h"

struct asset_signature *psignature;

/* ----------------------------------------------------------
 * FUNCTION	: read_sig_file
 * DESCRIPTION	: This function will read the signature file
 *		: into the signature data structure.
 * INPUT	: 0 - Data Structure
 * RETURN	: -1 - Error
 *		: 0 - Normal Return
 * ---------------------------------------------------------- */
int read_sig_file()
{
    FILE *fp;
    char line[300];
    char split1[3][300];
    char split2[4][100];
    int i, row, col;
    char title[300];
    const char *error;
    int erroffset;
    char sig_file[MAX_FILE];
    struct asset_signature *tmp;
    struct asset_signature *list;

    /* Check for a PADS_SIGNATURE_LIST file within the current directory.  */
    if ((fp = fopen(PADS_SIGNATURE_LIST, "r")) != NULL) {
	snprintf(sig_file, sizeof(sig_file), "./%s", PADS_SIGNATURE_LIST);
	fclose(fp);
    } else {
	snprintf(sig_file, sizeof(sig_file), "%s/%s", DATADIR, PADS_SIGNATURE_LIST);
    }

    /* Process File */
    if ((fp = fopen(sig_file, "r")) != NULL) {
	while (fgets(line, sizeof(line), fp) != NULL) {
	    i = 0; row = 0; col = 0;

	    /* Strip comments and skip if the line begins with ' '. */
	    strip_comment(line);
	    if (line[0] == ' ' || line[0] == '\n' || line[0] == '\r')
		continue;

	    /* Allocate memory for this signature. */
	    tmp = (struct asset_signature*)malloc(sizeof(struct asset_signature));

	    /* Strip out carriage returns. */
	    for (i = 0; i < sizeof(line); i++)
		if (line[i] == '\n' || line[i] == '\r')
		    line[i] = '\0';

	    /* Split line into 3 parts.  */
	    i = 0;
	    while (line[i] != '\0') {
		if (line[i] == ',' && row < 2) {
		    split1[row][col] = '\0';
		    row++;
		    col = 0;
		} else {
		    split1[row][col] = line[i];
		    col++;
		}

		i++;
	    }
	    split1[row][col] = '\0';

	    memcpy(tmp->service, split1[0], MAX_SERVICE);
	    tmp->regex = pcre_compile (split1[2], 0, &error, &erroffset, NULL);
	    if (tmp->regex == NULL) {
		printf("[!] Unable to compile signature:  %s\n", error);
		exit(0);
	    }

	    tmp->study = pcre_study (tmp->regex, 0, &error);
	    if (error != NULL) {
		printf("[!] Unable to study signature:  %s\n", error);
		exit(0);
	    }

	    /* Loop through title parsing as we go. */
	    strncpy(title, split1[1], sizeof(title));
	    row = 0;
	    col = 0;
	    i = 0;
	    while (title[i] != '\0' && row < 4 && col < 100) {
		if (title[i] == '/') {
		     split2[row][col] = '\0';
		     row++;
		     col = 0;
		} else {
		     split2[row][col] = title[i];
		     col++;
		}

		i++;
	     }
	     split2[row][col] = '\0';

	     /*
	     * Populate Signatures into Data Structure.
	     */
	    memcpy(tmp->title.app, split2[1], sizeof(char[100]));
	    memcpy(tmp->title.ver, split2[2], sizeof(char[25]));
	    memcpy(tmp->title.misc, split2[3], sizeof(char[100]));
	    tmp->next = NULL;

	    /* Attach 'tmp' to last record of psignature. */
	    if (psignature == NULL) {
		psignature = tmp;
	    } else {
		list = psignature;
		while (list != NULL) {
		    if (list->next == NULL) {
			list->next = tmp;
			break;
		    } else {
			list = list->next;
		    }
		}
	    }
	}

	fclose(fp);

    } else {
	fprintf(stderr, "[!] ERROR:  Unable to open %s!\n", sig_file);
	exit(0);
	return -1;
    }

}

/* ----------------------------------------------------------
 * FUNCTION	: tcp_ident
 * DESCRIPTION	: This function will take a TCP payload and
 *		: match it against the signature base.
 * INPUT	: 0 - IP Address
 *		: 1 - TCP Port
 *		: 2 - Packet Payload
 *		: 3 - Packet Payload Length
 * RETURN	: -1 - Error!
 *		:  0 - Normal Exit
 * ---------------------------------------------------------- */
int tcp_ident (struct in_addr ip_addr,
	       u_int16_t port,
	       char *payload,
	       int plen)
{
    unsigned short i_attempts;

    /* Retrieve i_attempts for this asset. */
    i_attempts = get_i_attempts(ip_addr, port, IPPROTO_TCP);

    if (i_attempts > 0) {
	i_attempts--;
	update_i_attempts(ip_addr, port, IPPROTO_TCP, i_attempts);

	if (pcre_match(ip_addr, port, IPPROTO_TCP, payload, plen) == 1) {
	    /* MATCH! */
	    i_attempts = 0;
	    update_i_attempts(ip_addr, port, IPPROTO_TCP, 0);
	}

	/* Print asset if this is the last time to identify it. */
	if (i_attempts == 0) {
	    print_asset(ip_addr, port, IPPROTO_TCP);
	}
    }

    return 0;
}

/* ----------------------------------------------------------
 * FUNCTION	: pcre_match
 * DESCRIPTION	: This function will match a string against
 *		: all the known signatures.
 * INPUT	: 0 - IP Address
 *		: 1 - Port
 *		: 2 - Proto
 *		: 3 - Payload
 *		: 4 - Payload Length
 * RETURN	: 0 - Not Matched
 *		: 1 - Matched
 * ---------------------------------------------------------- */
int pcre_match (struct in_addr ip_addr,
		u_int16_t port,
		unsigned short proto,
		const char *payload,
		int plen)
{
    struct asset_signature *list = psignature;
    int rc;
    int ovector[15];
    int i;
    char *app;

    while (list != NULL) {
	/* Execute Regular Expression */
	rc = pcre_exec(list->regex, list->study, payload, plen,
		0, 0, ovector, 15);

	if (rc != -1) {
	    app = get_app_name(list, payload, ovector, rc);
	    update_asset(ip_addr, port, proto, list->service, app);
	    free(app);
	    app = NULL;
	    return 1;
	}

	list = list->next;
    }

    return 0;
}

/* ----------------------------------------------------------
 * FUNCTION	: get_app_name
 * DESCRIPTION	: This function will take the results of a
 *		: pcre match and compile the application name
 *		: based off of the signature.
 * INPUT	: 0 - Signature Pointer
 *		: 1 - payload
 *		: 2 - ovector
 *		: 3 - rc (return from pcre_exec)
 * RETURN	: processed app name
 * ---------------------------------------------------------- */
char *get_app_name (struct asset_signature *sig,
		    const char *payload,
		    int *ovector,
		    int rc)
{
    char sub[100];
    char app[230];
    char expr[100];
    char *retval;
    int i = 0;
    int n = 0;
    int x = 0;
    int z = 0;

    /* Create Application string using the values in signature[i].title.  */
    if (strlen(sig->title.app) > 0) {
	strncpy(app, sig->title.app, MAX_APP);
    }
    if (strlen(sig->title.ver) > 0) {
	strcat(app, " ");
	strncat(app, sig->title.ver, MAX_VER);
    }
    if (strlen(sig->title.misc) > 0) {
	strcat(app, " (");
	strncat(app, sig->title.misc, MAX_MISC);
	strcat(app, ")");
    }

    /* Replace $1, $2, etc. with the appropriate substring.  */
    while (app[i] != '\0' && z < (sizeof(sub) - 1)) {
	/* Check to see if the string contains a $? mark variable. */
	if (app[i] == '$') {
	    /* Yes it does, replace it with the appropriate match string. */
	    i++;
	    n = atoi(&app[i]);

	    pcre_copy_substring(payload, ovector, rc, n, expr, sizeof(expr));
	    x = 0;
	    while (expr[x] != '\0' && z < (sizeof(sub) - 1)) {
		sub[z] = expr[x];
		z++;
		x++;
	    }
	    for (x = 0; x < sizeof(expr); x++)
		expr[x] = '\0';
	    i++;
	} else {
	    /* No it doesn't, copy to new string. */
	    sub[z] = app[i];
	    i++;
	    z++;
	}
    }
    sub[z] = '\0';

    retval = (char *) malloc(sizeof(sub));
    memcpy(retval, sub, sizeof(sub));
    return retval;

}

/* ----------------------------------------------------------
 * FUNCTION	: free_signature
 * DESCRIPTION	: This function will free the signatures
 *		: data structure from memory.
 * INPUT	: None!
 * RETURN	: None!
 * ---------------------------------------------------------- */
void free_signature ()
{
    struct asset_signature *next;

    /* Free records in psignature (asset_signature). */
    while (psignature != NULL) {
	next = psignature->next;
	free (psignature);
	psignature = next;
    }
}

/* ----------------------------------------------------------
 * FUNCTION	: print_signature
 * DESCRIPTION	: This function will print out all of the
 *		: signatures stored in the signature data
 *		: structure.
 * INPUT	: None!
 * RETURN	: None!
 * ---------------------------------------------------------- */
#ifdef DEBUG
void print_signature()
{
    struct asset_signature *list = psignature;
    int i = 1;

    while (list != NULL) {
	printf("[ %d ] ------------------\n", i);
	printf("1:  %s\n", list->service);
	printf("2a: %s\n", list->title.app);
	printf("2b: %s\n", list->title.ver);
	printf("2c: %s\n", list->title.misc);
	printf("3:  %s\n", list->regex);
	printf("\n");

	i++;
	list = list->next;
    }
}
#endif /* DEBUG */
