/*************************************************************************
 * asset-storage.c
 *
 * Matt Shelton <matt@mattshelton.com>
 *
 * This module contains functions related to the storage of assets.
 *
 * Copyright (C) 2004 Matt Shelton <matt@mattshelton.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: asset-storage.c,v 1.30 2004/09/30 22:25:50 mattshelton Exp $
 *
 **************************************************************************/
#include "asset-storage.h"

struct asset_storage *passet;
struct arp_storage *parp;
FILE *report;
#ifndef DISABLE_VENDOR
extern VENDOR_LINK vendors;
#endif /* DISABLE_VENDOR */

/* ----------------------------------------------------------
 * FUNCTION	: check_tcp_asset
 * DESCRIPTION	: This function determines whether an asset
 *		: has already been recorded.
 * INPUT	: 0 - IP Address
 *		: 1 - TCP port
 * RETURN	: 0 - Asset Exists
 *		: 1 - New Asset
 * ---------------------------------------------------------- */
int check_tcp_asset (struct in_addr ip_addr, u_int16_t port)
{
    struct asset_storage *rec;

    /* Check the Asset data structure for an existing entry. */
    rec = passet;
    while (rec != NULL) {
	if (rec->ip_addr.s_addr == ip_addr.s_addr
		&& rec->port == port
		&& rec->proto == IPPROTO_TCP) {
	    return 0;

	} else {
	    rec = rec->next;
	}
    }

    /* Asset not found, return 1 */
    return 1;
}

/* ----------------------------------------------------------
 * FUNCTION	: check_icmp_asset
 * DESCRIPTION	: This function determines whether an asset
 *		: has already been recorded.
 * INPUT	: 0 - IP Address
 * RETURN	: 0 - Asset Exists
 *		: 1 - New Asset
 * ---------------------------------------------------------- */
int check_icmp_asset (struct in_addr ip_addr)
{
    struct asset_storage *rec;

    /* Check the Asset data stucture for an existing entry. */
    rec = passet;
    while (rec != NULL) {
	if (rec->ip_addr.s_addr == ip_addr.s_addr
		&& rec->proto == IPPROTO_ICMP) {
	    return 0;

	} else {
	    rec = rec->next;
	}
    }

    /* Asset not found, return 1 */
    return 1;
}

/* ----------------------------------------------------------
 * FUNCTION	: check_arp_asset
 * DESCRIPTION	: This function determines whether the ARP
 *		: data for an asset has been recorded in the
 *		: ARP data structure.
 * INPUT	: 0 - IP Address
 *		: 1 - MAC Address
 * RETURN	: 0 - Asset Exists
 *		: 1 - New Asset
 * ---------------------------------------------------------- */
int check_arp_asset (struct in_addr ip_addr, struct ether_addr mac_addr)
{
    struct arp_storage *rec;

    /* Check the ARP data structure for an existing entry. */
    rec = parp;
    while (rec != NULL) {
	if (rec->ip_addr.s_addr == ip_addr.s_addr
		&& (strcmp(MAC_ADDR(rec->mac_addr),MAC_ADDR(mac_addr)) == 0)) {
	    return 0;

	} else {
	    rec = rec->next;
	}
    }

    /* Asset not found return 1 */
    return 1;
}

/* ----------------------------------------------------------
 * FUNCTION	: add_asset
 * DESCRIPTION	: This function will add an asset to the
 *		: specified asset data structure.
 * INPUT	: 0 - IP Address
 *		: 1 - Port
 *		: 2 - Protocol
 *		: 3 - Service
 *		: 4 - Application
 *		: 5 - Discovered
 * RETURN	: None!
 * ---------------------------------------------------------- */
void add_asset (struct in_addr ip_addr,
		u_int16_t port,
		unsigned short proto,
		char service[MAX_SERVICE],
		char application[MAX_LENGTH],
		time_t discovered)
{
    struct asset_storage *tmp;
    struct asset_storage *data;

    /* Assign data to temp structure.  */
    tmp = (struct asset_storage*)malloc(sizeof(struct asset_storage));
    tmp->ip_addr.s_addr = ip_addr.s_addr;
    tmp->port = port;
    tmp->proto = proto;
    memcpy(tmp->service, service, MAX_SERVICE);
    memcpy(tmp->application, application, MAX_LENGTH);
    tmp->next = NULL;

    /*
     * If this device has been read from a report file, set
     * the discovered time to whatever is in the report.
     * Also, we don't want to try to identify this service
     * anymore (i_attempts = 0);
     */
    if (!discovered) {
	tmp->discovered = time(NULL);
	tmp->i_attempts = I_ATTEMPTS;
    } else {
	tmp->discovered = discovered;
	tmp->i_attempts = 0;
    }

    /*
     * ICMP packets will not be identified, set i_attempts
     * to zero.
     */
    if (proto == IPPROTO_ICMP) {
	tmp->i_attempts = 0;
    }

    /* Find this record's location within linked list.  */
    if (passet == NULL) {
	passet = tmp;
    } else {
	data = passet;
	while (data != NULL) {
	    if (data->next == NULL) {
		data->next = tmp;
		break;
	    } else {
		data = data->next;
	    }
	}
    }

    return;
}

/* ----------------------------------------------------------
 * FUNCTION	: add_arp_asset
 * DESCRIPTION	: This function will add an ARP entry to the
 *		: ARP data structure.
 * INPUT	: 0 - IP Address
 *		: 1 - MAC Address
 *		: 2 - Discovered
 * RETURN	: None!
 * ---------------------------------------------------------- */
void add_arp_asset (struct in_addr ip_addr, struct ether_addr mac_addr,
		    time_t discovered)
{
    struct arp_storage *data;
    struct arp_storage *tmp;

    tmp = (struct arp_storage*)malloc(sizeof(struct arp_storage));
    tmp->ip_addr.s_addr = ip_addr.s_addr;
    memcpy(MAC_ADDR(&tmp->mac_addr), MAC_ADDR(mac_addr), sizeof(u_int8_t) * 6);
    tmp->next = NULL;

    /*
     * If this device has been read from a report file, set
     * the discovered time to whatever is in the report.
     * Also, we don't want to try to identify this service
     * anymore (i_attempts = 0);
     */
    if (!discovered) {
	tmp->discovered = time(NULL);
    } else {
	tmp->discovered = discovered;
    }

    /* Find this record's location within linked list.  */
    if (parp == NULL) {
	parp = tmp;
    } else {
	data = parp;

	while (data != NULL) {
	    if (data->next == NULL) {
		data->next = tmp;
		break;
	    } else {
		data = data->next;
	    }
	}
    }
}

/* ----------------------------------------------------------
 * FUNCTION	: print_asset
 * DESCRIPTION	: This function will print the specified
 *		: asset to the screen / output file.
 * INPUT	: 0 - IP Address
 *		: 1 - Port
 *		: 2 - Proto
 * RETURN	: 0 - Failed!
 *		: 1 - Success!
 * ---------------------------------------------------------- */
short print_asset (struct in_addr ip_addr, u_int16_t port, unsigned short proto)
{
    struct asset_storage *list;
    struct asset_storage *rec;

    /* Retrieve record from linked list.  */
    list = passet;
    while (list != NULL) {
	if (ip_addr.s_addr == list->ip_addr.s_addr
		&& port == list->port
		&& proto == list->proto) {
	    /* Found! */
	    rec = list;
	    break;
	} else {
	    list = list->next;
	}
    }

    /* Make sure that a record was found. */
    if (rec == NULL)
	return 0;

    /* Print to Screen */
    fprintf(stdout, "[*] Asset Found:  Port - %d / Host - %s / Service - %s / Application - %s\n",
		ntohs(rec->port), inet_ntoa(rec->ip_addr),
		rec->service, rec->application);

    /* Print to File */
    if (report != NULL) {
	fprintf(report, "%s,%d,%d,%s,%s,%d\n",
		inet_ntoa(rec->ip_addr), ntohs(rec->port),
		rec->proto, rec->service,
		rec->application, rec->discovered);
	fflush(report);
    } else {
	fprintf(stderr, "[!] ERROR:  File handle not open!\n");
    }

    return 1;
}

/* ----------------------------------------------------------
 * FUNCTION	: print_arp_asset
 * DESCRIPTION	: This function will print out the ARP asset
 *		: to the screen and to the report file.
 * INPUT	: 0 - IP Address
 *		: 1 - MAC Address
 * RETURN	: 0 - Failed!
 *		: 1 - Success!
 * ---------------------------------------------------------- */
short print_arp_asset (struct in_addr ip_addr, struct ether_addr mac_addr)
{
    struct arp_storage *list;
    struct arp_storage *rec;
#ifndef DISABLE_VENDOR
    char *vendor = NULL;
#endif

    /* Retrieve record from linked list.  */
    list = parp;
    while (list != NULL) {
	if (ip_addr.s_addr == list->ip_addr.s_addr
		&& (strcmp(MAC_ADDR(mac_addr), MAC_ADDR(list->mac_addr)) == 0)) {

	    /* Found! */
	    rec = list;
	    break;
	} else {
	    list = list->next;
	}
    }

    /* Make sure that a record was found. */
    if (list == NULL || rec == NULL)
	return 0;

    /* Print to Screen */
#ifndef DISABLE_VENDOR
    vendor = get_vendor(vendors, MAC_ADDR(rec->mac_addr));
    if(vendor != NULL){
	fprintf(stdout, "[*] Asset Found:  IP Address - %s / MAC Address - %s (%s)\n", inet_ntoa(rec->ip_addr), ether_ntoa(&rec->mac_addr), vendor);
    } else {
	fprintf(stdout, "[*] Asset Found:  IP Address - %s / MAC Address - %s\n", inet_ntoa(rec->ip_addr), ether_ntoa(&rec->mac_addr));
    }
#endif /* DISABLE_VENDOR */
#ifdef DISABLE_VENDOR
    fprintf(stdout, "[*] Asset Found:  IP Address - %s / MAC Address - %s\n", inet_ntoa(rec->ip_addr), ether_ntoa(&rec->mac_addr));
#endif /* ENABLE_VENDOR */

     /* Print to File */
    if (report != NULL) {
#ifndef DISABLE_VENDOR
	if (vendor != NULL) {
	    fprintf(report, "%s,0,0,ARP (%s),%s,%d\n", inet_ntoa(rec->ip_addr),
		    vendor, ether_ntoa(&rec->mac_addr), rec->discovered);
	} else {
	    fprintf(report, "%s,0,0,ARP,%s,%d\n", inet_ntoa(rec->ip_addr),
		    ether_ntoa(&rec->mac_addr), rec->discovered);
	}
#endif /* DISABLE_VENDOR */
#ifdef DISABLE_VENDOR
	fprintf(report, "%s,0,0,ARP,%s,%d\n", inet_ntoa(rec->ip_addr), ether_ntoa(&rec->mac_addr), rec->discovered);
#endif /* ENABLE_VENDOR */

	fflush(report);
    } else {
	fprintf(stderr, "[!] ERROR:  File handle not open!\n");
    }

    return 1;
}

/* ----------------------------------------------------------
 * FUNCTION	: init_output_file
 * DESCRIPTION	: This function will initialize the output
 *		: CSV file.  If the file already exists, it
 *		: will read in the file and add each asset
 *		: to the asset data structure.
 * INPUT	: 0 - CSV filename
 * RETURN	: None!
 * --------------------------------------------------------- */
void init_output_file (char outfile[255])
{
    FILE *fp;

    /* Check to see if *outfile exists. */
    if ((fp = fopen(outfile, "r")) == NULL) {

	/* File does not exist, create new.. */
	if ((report = fopen(outfile, "w")) != NULL) {
	    fprintf(report, "asset,port,proto,service,application,discovered\n");

	} else {
	    fprintf(stderr, "[!] ERROR:  Cannot open file %s!\n", outfile);
	    exit(-1);
	}

    } else {

	/* File does exist, read it into data structure. */
	fclose(fp);
	read_report_file(outfile);

	/* Open file and assign it to the global FILE pointer.  */
	if ((report = fopen(outfile, "a")) == NULL) {
	    fprintf(stderr, "[!] ERROR:  Cannot open file %s!\n", outfile);
	    exit(-1);
	}
    }

    return;
}

/* ----------------------------------------------------------
 * FUNCTION	: close_output_file
 * DESCRIPTION	: This function will close the report file
 *		: defined in global variable FILE *report.
 * INPUT	: None!
 * RETURN	: None!
 * ---------------------------------------------------------- */
void close_output_file ()
{
    fclose(report);
}

/* ----------------------------------------------------------
 * FUNCTION	: read_report_file
 * DESCRIPTION	: This function will read in a specified
 *		: report CSV file.  It will then break a part
 *		: the line and add the assets to the
 *		: specified asset data structure.
 * INPUT	: 0 - CSV Filename
 * RETURN	: None!
 * ---------------------------------------------------------- */
void read_report_file (char infile[255])
{
    FILE *fp;
    char line[255];
    char split[MAX_LENGTH];
    int i, x, col;

    /* Temporary Storage */
    char tmp_addr[16];
    struct in_addr ip_addr;
    struct ether_addr *mac_addr;
    int port;
    int proto;
    char service[MAX_SERVICE];
    char application[MAX_LENGTH];
    time_t discovered;

    printf("[-] Processing Existing %s\n", infile);

    /* Process File */
    if ((fp = fopen(infile, "r")) != NULL) {
	while (fgets(line, sizeof(line), fp) != NULL) {
	    /* Make sure that this isn't the title line. */
	    if (!isdigit(line[0]))
		continue;

	    i = 0; x = 0; col = 0;

	    /* Split Line Apart */
	    while (line[i] != '\0' && i < sizeof(line)) {
		if (line[i] == ',' || line[i] == '\n') {
		    split[x] = '\0';

		    switch (col) {
			case 0:
			    if ((inet_aton(&split, &ip_addr)) == -1)
				continue;
			    break;

			case 1:
			    if ((port = htons(atoi((split)))) == -1)
				continue;
			    break;

			case 2:
			    if ((proto = atoi(split)) == -1)
				continue;
			    break;

			case 3:
			    if ((strncpy(service, split, MAX_SERVICE)) == NULL)
				continue;
			    break;

			case 4:
			    if ((strncpy(application, split, MAX_LENGTH)) == NULL)
				continue;
			    break;

			case 5:
			    if ((discovered = atol(split)) == -1)
				continue;
			    break;
		    }

		    col++;
		    x = 0;
		    split[0] = '\0';

		} else {
		    split[x] = line[i];
		    x++;
		}

		i++;
	    }

	    /* Make sure that this line contains 'good' data. */
	    if (strlen(service) == 0 || strlen(application) == 0
		    || discovered <= 0)
		continue;

	    /* Add Asset to Data Structure */
	    if (proto == 0) {
		/* ARP */
		mac_addr = ether_aton(application);
		add_arp_asset(ip_addr, *mac_addr, discovered);
	    } else {
		/* Everything Else */
		add_asset(ip_addr, port, proto, service, application, discovered);
	    }
	}

	fclose(fp);
    }
    else
	fprintf(stderr, "[!] ERROR:  Unable to open %s!\n", infile);

}

/* ----------------------------------------------------------
 * FUNCTION	: get_i_attempts
 * DESCRIPTION	: Return an asset's i_attempts value.
 * INPUT	: 0 - IP Address
 *		: 1 - Port
 *		: 2 - Proto
 * RETURN	: unsigned short i_attempts;
 * ---------------------------------------------------------- */
unsigned short get_i_attempts (struct in_addr ip_addr,
			       u_int16_t port,
			       unsigned short proto)
{
    struct asset_storage *rec;

    /* Find asset within linked list.  */
    rec = passet;
    while (rec != NULL) {
	if (ip_addr.s_addr == rec->ip_addr.s_addr
		&& port == rec->port
		&& proto == rec->proto) {
	    /* Found! */
	    return rec->i_attempts;

	} else {
	    rec = rec->next;
	}
    }

    return 0;
}

/* ----------------------------------------------------------
 * FUNCTION	: update_i_attempts
 * DESCRIPTION	: Updates the i_attempts field for a
 *		: specified asset.
 * INPUT	: 0 - IP Address
 *		: 1 - Port
 *		: 2 - Proto
 *		: 3 - i_attempts
 * RETURN	: 0 - Success
 *		: 1 - Failure
 * ---------------------------------------------------------- */
short update_i_attempts (struct in_addr ip_addr,
			 u_int16_t port,
			 unsigned short proto,
			 unsigned short i_attempts)
{
    struct asset_storage *rec;

    /* Find asset within linked list.  */
    rec = passet;
    while (rec != NULL) {
	if (ip_addr.s_addr == rec->ip_addr.s_addr
		&& port == rec->port
		&& proto == rec->proto) {
	    /* Found! */
	    rec->i_attempts = i_attempts;
	    return 0;

	} else {
	    rec = rec->next;
	}
    }

    return 1;
}

/* ----------------------------------------------------------
 * FUNCTION	: update_asset
 * DESCRIPTION	: This function will update the service and
 *		: application fields of an asset.
 * INPUT	: 0 - IP Address
 *		: 1 - Port
 *		: 2 - Proto
 *		: 3 - Service
 *		: 4 - Application
 * RETURN	: 0 - Success!
 *		: 1 - Failure!
 * ---------------------------------------------------------- */
short update_asset (struct in_addr ip_addr,
		    u_int16_t port,
		    unsigned short proto,
		    char *service,
		    char *application)
{
    struct asset_storage *rec;
    /* Find asset within linked list.  */
    rec = passet;
    while (rec != NULL) {
	if (ip_addr.s_addr == rec->ip_addr.s_addr
		&& port == rec->port
		&& proto == rec->proto) {
	    /* Found! */
	    strncpy(rec->service, service, MAX_SERVICE);
	    strncpy(rec->application, application, MAX_LENGTH);
	    return 0;

	} else {
	    rec = rec->next;
	}
    }

    return 1;
}

/* ----------------------------------------------------------
 * FUNCTION	: free_asset
 * DESCRIPTION	: This function will free all the records
 *		: placed in the asset data structure.
 * INPUT	: None!
 * RETURN	: None!
 * ---------------------------------------------------------- */
void free_asset ()
{
    struct asset_storage *next1;
    struct arp_storage *next2;

    /* Free records in passet (asset_storage). */
    while (passet != NULL) {
	next1 = passet->next;
	free (passet);
	passet = next1;
    }

    /* Free records in parp (arp_storage). */
    while (parp != NULL) {
	next2 = parp->next;
	free (parp);
	parp = next2;
    }
}

/* ----------------------------------------------------------
 * FUNCTION	: print_database
 * DESCRIPTION	: This function prints out the asset data
 *		: structure.  It is mainly for debug
 *		: purposes.
 * INPUT	: None!
 * RETURN	: None!
 * ---------------------------------------------------------- */
#ifdef DEBUG
void print_database ()
{
    struct asset_storage *rec;
    struct arp_storage *arp;
    int id = 0;

    printf("-- Begin Asset Database --\n");
    rec = passet;
    while (rec != NULL) {
	printf("%d:  %s,%d,%d,%d,%s,%s,%d\n",
		id, inet_ntoa(rec->ip_addr), ntohs(rec->port),
		rec->proto, rec->discovered,
		rec->service, rec->application,
		rec->i_attempts);
	rec = rec->next;
	id++;
    }
    printf("-- End Asset Database --\n\n");

    printf("-- Begin ARP Database --\n");
    id = 0;
    arp = parp;
    while (arp != NULL) {
	printf("%d:  %s,%s,%d\n", id, inet_ntoa(arp->ip_addr),
		ether_ntoa(&arp->mac_addr), arp->discovered);
	arp = arp->next;
	id++;
    }
    printf("-- End ARP Database --\n\n");
}
#endif /* DEBUG */

