/*************************************************************************
 * pads.c
 *
 * Matt Shelton	<matt@mattshelton.com>
 *
 * The purpose of this system is to determine network assets by passively
 * listening to network traffic.
 *
 * Copyright (C) 2004 Matt Shelton <matt@mattshelton.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: pads.c,v 1.40 2004/09/30 22:25:50 mattshelton Exp $
 *
 **************************************************************************/
#include "pads.h"

/* Variable Declarations */
pcap_t *handle;
char *dev = NULL;
char errbuf[PCAP_ERRBUF_SIZE];
proc_t processor;
char report_file[255] = "assets.csv";
#ifndef DISABLE_VENDOR
extern VENDOR_LINK vendors;
#endif /* DISABLE_VENDOR */

/* ----------------------------------------------------------
 * FUNCTION	: process_pkt
 * DESCRIPTION	: This function takes data from libpcap and
 *		: processes it based on what type of traffic
 *		: it is.
 * INPUT	: 0 - Not used
 *		: 1 - pcap packet header
 *		: 2 - pcap packet
 * ---------------------------------------------------------- */
void process_pkt (u_char *args, const struct pcap_pkthdr* pkthdr, const u_char* packet)
{
    /* Call LLC Processor */
    (*processor)(pkthdr, packet);
}

/* ----------------------------------------------------------
 * FUNCTION	: set_processor
 * DESCRIPTION	: This function determines what type of LLC
 *		: type is being used.  It will then set a
 *		: pointer to the correct LLC function.
 * INPUT	: PCAP Handle
 * RETURN	: None!
 * ---------------------------------------------------------- */
void set_processor (pcap_t *this_handle)
{
    int datalink;
    datalink = pcap_datalink(this_handle);

    switch (datalink) {
	/* Ethernet */
	case DLT_EN10MB:
	    processor = process_eth;
	    break;

#ifdef DLT_LINUX_SLL
	/* Linux Cooked Sockets */
	case DLT_LINUX_SLL:
	    processor = process_sll;
	    break;
#endif /* DLT_LINUX_SLL */

	/* Default */
	default:
	    printf("[!] ERROR:  LLC not supported!  Please contact the author!\n");
	    die_clean();
	    break;
    }

    return;
}

/* ----------------------------------------------------------
 * FUNCTION	: print_header
 * DESCRIPTION	: Prints initial header.
 * ---------------------------------------------------------- */
void print_header ()
{
    printf("pads - Passive Asset Detection System\n");
    printf("v%s - %s\n", PADS_VERSION, PADS_DATE);
    printf("Matt Shelton <matt@mattshelton.com>\n");
    printf("\n");
}

/* ----------------------------------------------------------
 * FUNCTION	: print_usage
 * DESCRIPTION	: Prints the Program Usage
 * ---------------------------------------------------------- */
void print_usage()
{
    printf("Usage:\n"
	   "-i <interface>		: Listen on <interface>.  The lowest number interface\n"
	   "			  will be used if an interface isn't specified.\n"
	   "-r <file>		: Read packets from a libpcap formatted file.\n"
	   "-w <file>		: Dump data into file other than assets.csv.\n"
	   "-n <network>		: Reads in a comma seperated list of networks\n"
	   "			  to be monitored.\n"
	   "				ex.  -n \"192.168.0.0/24,10.0.0.0/16\"\n"
	   "-D			: Run PADS in the background (daemon mode).\n"
	   "-h			: Help\n"
	   "\n"
	   "Additional arguments will be processed as a libpcap filter.  For example,\n"
	   "the following command will not only use interface hme1 but will also only\n"
	   "search for assets on port 22:\n"
	   "\n"
	   "	pads -i hme1 port 22\n");
    printf("\n");
    exit(0);
}

/* ----------------------------------------------------------
 * FUNCTION	: die_clean
 * DESCRIPTION	: This function needs to be called before the
 *		: program exits.  It shuts down libpcap and
 *		: everything.
 * ---------------------------------------------------------- */
void die_clean()
{
    struct pcap_stat pstat;

    /* Display PCAP Statistics */
    if (!pcap_stats(handle, &pstat))
    {
	printf("\n");
	printf("[*] %d Packets Received\n", pstat.ps_recv);
	printf("[*] %d Packets Dropped by Software\n", pstat.ps_drop);
	printf("[*] %d Packets Dropped by Interface\n", pstat.ps_ifdrop);
	printf("\n");
    }

    /* print_database(); */  /* Debug */

    close_output_file();

    /* Kill PCAP Object */
    if (handle)
    {
	printf("[-] Closing PCAP Connection\n");
	pcap_close(handle);
    }

    /* Free Data Structures */
    free_asset();
    free_signature();
#ifndef DISABLE_VENDOR
    free_vendor(vendors);
#endif

    exit(0);
}



/* ----------------------------------------------------------
 * FUNCTION	: main
 * ---------------------------------------------------------- */
int main(int argc, char *argv[])
{
    /* Variables */
    int ch;
    int i;
    struct pcap_pkthdr header;		/* The header that pcap gives us */
    bpf_u_int32 mask;			/* The netmask of our sniffing device */
    bpf_u_int32 net;			/* The IP of our sniffing device */
    const u_char *packet;		/* The actual packet */
    char *pcap_file = NULL;		/* PCAP file used only if '-r' switch specified. */
    char *pcap_filter;			/* PCAP filter text (specified on command line). */
    struct bpf_program filter;		/* PCAP filter structure */

    print_header();

    /* Signal Trapping */
    signal(SIGINT, die_clean);
    signal(SIGQUIT, die_clean);
    signal(SIGTERM, die_clean);
    signal(SIGABRT, die_clean);
    signal(SIGPIPE, die_clean);

    /* Process Command Line Arguments */
    while ((ch = getopt(argc, argv, "hDi:w:r:an:")) != -1)  {
	switch (ch) {
	    case 'h':
		print_usage();
		break;
	    case 'D':
		daemonize();
		break;
	    case 'i':
		dev = optarg;
		break;
	    case 'a':
		/* Do nothing for now. */
		break;
	    case 'r':
		pcap_file = optarg;
		break;
	    case 'w':
		strncpy(report_file, optarg, sizeof(report_file) - 1);
		break;
	    case 'n':
		parse_networks(optarg);
		break;
	    default:
		print_usage();
		break;
	}
    }

    argc -= optind;
    argv += optind;

    /* Parse the rest of the command line. */
    if (argc > 0) {
	pcap_filter = copy_argv(argv);
    }

    /* Initialize Report File */
    init_output_file(report_file);

    /* Initialize Signature File */
    if (read_sig_file() == -1)
	exit(0);

#ifndef DISABLE_VENDOR
    /* Initialize MAC Vendor File */
    init_vendor();
#endif

    if (pcap_file) {
	/* Read from PCAP file specified by '-r' switch. */
	printf("[-] Reading from file %s\n", pcap_file);
	if (!(handle = pcap_open_offline(pcap_file, errbuf))) {
	    fprintf(stderr, "[!] ERROR:  Unable to open %s.  (%s)\n", pcap_file, errbuf);
	    exit(0);
	}

	mask = 0;
	net = 0;

    } else {
	/* Only root can access the interface. */
	if(geteuid() != 0) {
	    fprintf(stderr, "[!] ERROR:  Must be root!\n");
	    exit(0);
	}

	/* Determine Sniffing Interface */
	if (!dev)
	    if (!(dev = pcap_lookupdev(errbuf))) {
		fprintf(stderr, "[!] ERROR:  Unable to find a sniffing interface!  (%s)\n", errbuf);
		exit(0);
	    }

	/* Set up libpcap connection. */
	printf("[-] Listening on interface %s\n", dev);
	handle = pcap_open_live(dev, BUFSIZ, 1, 0, errbuf);

	/* Lookup Netmask (used with pcap_setfilter) */
	if (pcap_lookupnet(dev, &net, &mask, errbuf) == -1) {
	    fprintf(stderr, "[!] ERROR:  pcap_lookupnet (%s)\n", errbuf);
	    memset(&net, 0, sizeof(net));
	    memset(&mask, 0, sizeof(mask));
	}

    }

    /* Determine LLC Type */
    set_processor(handle);

    /* Compile libpcap filter */
    if (argc > 0) {
	printf("[-] Filter:  %s\n", pcap_filter);
	if (pcap_compile(handle, &filter, pcap_filter, 0, net) == -1) {
	    pcap_perror(handle, "[!] ERROR:  Unable to compile pcap filter!  (syntax error)\n");
	    die_clean();
	}
	if (pcap_setfilter(handle, &filter)) {
	    pcap_perror(handle, "[!] ERROR:  Unable to set pcap filter!\n");
	    die_clean();
	}
	free(pcap_filter);
    }

    /* Sniff libpcap connection. */
    printf("\n");
    while (pcap_loop(handle, -1, process_pkt, NULL));

    die_clean();
    return(0);
}

